/**
 * MultipartStripper.java
 * Created on 2003/06/05 
 * Copyright (c) by es , All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.prixdami.alexis.multipart;

import java.io.*;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.*;

import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServletRequest;

/**
 * form-Muitipart data splitter
 * @author es
 * @version 2.00
 */
public class MultipartStripper {
    private static final int BYTE_TO_ARRAY = 0;
    private static final int BYTE_TO_FILE = 1;
    private static final String BOUNDARY = "boundary=";
    private static final String CONTENT = "Content-Disposition:";
    private static final String FILENAME = "filename=\"";
    private static final String NAME = "name=\"";
    private static final String CONTENTTYPE = "Content-Type";
    private static final String MULTIPART = "multipart/";

    private String _tempPath = "./";
    private String _encoding = null;
    private File _tempDir = null;
	private MessageDigest _md;
    private boolean _eof2abort = true;

    /**
     * @param request
     * @return
     */
    public static boolean isMultipartContent(HttpServletRequest request) {
        String contentType = request.getHeader(CONTENTTYPE);
        if (contentType == null) {
            return false;
        }
        if (contentType.startsWith(MULTIPART)) {
            return true;
        }
        return false;
    }

    /**
     * 
     */
    public MultipartStripper() {
        init();
    }

    /**
     * @param encoding
     * @param path
     */
    public MultipartStripper(String encoding, String path) {
        this._encoding = encoding;
        this._tempPath = path;
        init();
    }

    /**
     * @param path
     */
    public MultipartStripper(String path) {
        this._tempPath = path;
        init();
    }

    /**
     * 
     */
    private void init() {
        _tempDir = new File(_tempPath);
        _tempDir.mkdir();
        try {
            _md = MessageDigest.getInstance("MD5");
        } catch (NoSuchAlgorithmException e) {
            System.err.println("MultipartStripper : Warning! can't use MD5-Digest");
        }
    }

    public void setAbortMode(boolean flag) {
        _eof2abort = flag;
    }

    public Map strip(ServletRequest request) throws IOException {
        return strip(request, -1);
    }

    public Map strip(ServletRequest request, long maxsize) throws IOException {
        return strip(request.getInputStream(), request.getContentType(), request.getCharacterEncoding(), request.getContentLength(), maxsize);
    }

    public Map strip(InputStream inputStream, String contentType, long contentLength) throws IOException {
        return strip(inputStream, contentType, null, contentLength, -1);
    }

    public Map strip(InputStream inputStream, String contentType, String charset, long contentLength, long limitSize) throws IOException {
		if (contentLength > limitSize && limitSize > 0) {
			throw new LimitSizeOverFlow("request is too huge :" + contentLength);
		}

		String encode = (charset == null) ? _encoding : charset;
        MultipartInputStream in = new MultipartInputStream(inputStream);
        Map returnValue = new HashMap();
        boolean eof = false;
        byte[] buffer = new byte[512];

        String boundary = getBoundary(contentType);
        if (boundary == null) {
            throw new MultipartParseException("this data is not MultiPart request.");
        }

        String currentLine;
        while (!eof) {
            currentLine = getLine(in, buffer, encode);
            if (currentLine == null) {
                eof = true;
                continue;
            }

            if (currentLine.indexOf(boundary) != -1) {
                currentLine = getLine(in, buffer, encode);
                if (currentLine == null) {
                    eof = true;
                    continue;
                }
            }

            if (currentLine.lastIndexOf(CONTENT) == -1) {
                throw new MultipartParseException("can't strip request.[Content-Disposition]");
            }

            String paramName = getToken(currentLine, NAME);
            String filename = getToken(currentLine, FILENAME);

            try {
                if (filename == null) {
                    //Normal-parameter
                    getLine(in, buffer, encode); //skip 1-line
                    byte[] rawParam = (byte[]) getBytes(in, boundary, BYTE_TO_ARRAY);

                    String param;
                    if (encode == null) {
                        param = new String(rawParam, 0, rawParam.length);
                    } else {
                        param = new String(rawParam, 0, rawParam.length, encode);
                    }

                    List lst = (List) returnValue.get(paramName);
                    if (lst == null) {
                        lst = new ArrayList();
                    }
                    lst.add(param);
                    returnValue.put(paramName, lst);
                } else if (filename.equals("")) {
                    //empty file-parameter
                    for (int i = 0; i < 3; i++) {
                        getLine(in, buffer, encode); //skip 3-line
                    }
                } else {
                    //file-parameter
                    String mimeType = getMimeType(getLine(in, buffer, encode));
                    getLine(in, buffer, encode); //skip 1-line
                    File rawFile = (File) getBytes(in, boundary, BYTE_TO_FILE);
                    MultipartFile data = new BasementMultipartFile(paramName, filename, mimeType, rawFile);
                    returnValue.put(paramName, data);
                }

            } catch (EOFException ex) {
                if (_eof2abort) {
                    throw ex;
                }
            } catch (Exception ex) {
                for (Iterator i = returnValue.keySet().iterator(); i.hasNext();) {
                    Object o = returnValue.get(i.next());
                    if (!(o instanceof BasementMultipartFile)) {
                        continue;
                    }
                    ((BasementMultipartFile) o).killFile();
                }

                if (ex instanceof IOException) {
                    throw (IOException) ex;
                }
                throw new MultipartParseException(ex.getMessage());
            }

        }
        return returnValue;
    }

    private Object getBytes(MultipartInputStream in, String boundary, int flag) throws IOException {
        OutputStream out = null;
        File f = null;

        switch (flag) {
            case BYTE_TO_ARRAY :
                out = new ByteArrayOutputStream();
                break;
            case BYTE_TO_FILE :
                f = new File(_tempDir, makeTempFileName());
                out = new BufferedOutputStream(new FileOutputStream(f));
                break;
            default :
                throw new IllegalArgumentException("wrong data store flag :" + flag);
        }

        try {
            fillByte(in, out, boundary);
            switch (flag) {
                case BYTE_TO_ARRAY :
                    int size = ((ByteArrayOutputStream) out).size();
                    byte[] returnValue = new byte[size];
                    System.arraycopy(((ByteArrayOutputStream) out).toByteArray(), 0, returnValue, 0, size);
                    return returnValue;
                case BYTE_TO_FILE :
                    out.flush();
                    return f;
                default :
                    throw new IllegalArgumentException("wrong data store flag :" + flag);
            }
        } catch (IOException ex) {
            if (f != null) {
                f.delete();
            }
            throw ex;
        } finally {
            out.close();
        }

    }

    private void fillByte(MultipartInputStream src, OutputStream dst, String boundary) throws IOException {
        byte[] buffer = new byte[4096 * 4];
        int maxRead = buffer.length - boundary.length() - 2;
        boolean eof = false;
        boolean abort = false;
        int fullcount = 0;

        while (!eof) {
            int readCount = 0;
            int currentPos = 0;

            do {
                readCount = src.readLine(buffer, currentPos, maxRead - currentPos);
                if (readCount == -1) {
                    eof = true;
                    abort = true;
                }
                if (readCount >= boundary.length()) {
                    eof = true;
                    for (int i = 0; i < boundary.length(); i++) {
                        if (boundary.charAt(i) != buffer[currentPos + i]) {
                            eof = false;
                            break;
                        }
                    }
                }
                if (!eof) {
                    currentPos += readCount;
                    fullcount += readCount;
                } else {
                    currentPos -= 2;
                }
            } while (!eof && maxRead > currentPos);

            if (abort) {
                currentPos -= 2;
                for (int i = 1; i < currentPos; i++) {
                    if (boundary.charAt(boundary.length() - i) == buffer[currentPos - i]) {
                        abort = false;
                    } else {
                        abort = true;
                        break;
                    }
                }
            } else {
                dst.write(buffer, 0, currentPos);
            }
        }

        if (abort) {
            throw new EOFException("EOF found.");
        }
    }

    private String getLine(MultipartInputStream in, byte[] buffer, String encode) throws java.io.IOException {
        int currentLength = in.readLine(buffer, 0, buffer.length);

        if (currentLength == -1) {
            return null;
        }

        String line;
        if (encode != null) {
            line = new String(buffer, 0, currentLength - 2, encode);
        } else {
            line = new String(buffer, 0, currentLength - 2);
        }
        return line;
    }

    private String getBoundary(String line) {
        int index = line.lastIndexOf(BOUNDARY);

        if (index == -1)
            return null;
        String boundary = line.substring(index + BOUNDARY.length());
        boundary = "--" + boundary;
        return boundary;
    }

    private String getMimeType(String line) {
        int index = line.lastIndexOf(CONTENTTYPE);

        if (index == -1) {
            return null;
        }
        String attr = line.substring(index + CONTENTTYPE.length() + 2);
        return attr;
    }

    private String getToken(String line, String key) {
        String[] params = line.split(";");

        for (int i = 0; i < params.length; i++) {
            if (params[i].indexOf(key) != -1) {
                line = params[i];
                break;
            }
        }
        int startpos = line.lastIndexOf(key);
        int endpos = line.lastIndexOf("\"");
        if (startpos == -1) {
            return null;
        }
        String attr = line.substring(startpos + key.length(), endpos);
        return attr;
    }

    private String makeTempFileName() {
        String time = String.valueOf(System.currentTimeMillis());
        if (_md == null) {
            return time;
        }

        StringBuffer sb = new StringBuffer();
        byte[] data = _md.digest(time.getBytes());
        for (int i = 0; i < data.length; i++) {
            sb.append(Integer.toHexString(data[i] & 0x00ff));
        }
        return sb.toString();
    }

}